from inquirer3.prompt import prompt
from inquirer3.questions import Checkbox
from inquirer3.questions import Confirm
from inquirer3.questions import Editor
from inquirer3.questions import List
from inquirer3.questions import Password
from inquirer3.questions import Path
from inquirer3.questions import Text
from inquirer3.questions import load_from_dict
from inquirer3.questions import load_from_json
from inquirer3.questions import load_from_list
from inquirer3.shortcuts import checkbox
from inquirer3.shortcuts import confirm
from inquirer3.shortcuts import editor
from inquirer3.shortcuts import list_input
from inquirer3.shortcuts import password
from inquirer3.shortcuts import text


__all__ = [
    "prompt",
    "Text",
    "Editor",
    "Password",
    "Confirm",
    "List",
    "Checkbox",
    "Path",
    "load_from_list",
    "load_from_dict",
    "load_from_json",
    "text",
    "editor",
    "password",
    "confirm",
    "list_input",
    "checkbox",
]



from readchar import key

from inquirer3 import errors
from inquirer3.render.console._other import GLOBAL_OTHER_CHOICE
from inquirer3.render.console.base import MAX_OPTIONS_DISPLAYED_AT_ONCE
from inquirer3.render.console.base import BaseConsoleRender
from inquirer3.render.console.base import half_options


class List(BaseConsoleRender):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current = self._current_index()

    @property
    def is_long(self):
        choices = self.question.choices or []
        return len(choices) >= MAX_OPTIONS_DISPLAYED_AT_ONCE

    def get_options(self):
        choices = self.question.choices or []
        if self.is_long:
            cmin = 0
            cmax = MAX_OPTIONS_DISPLAYED_AT_ONCE

            if half_options < self.current < len(choices) - half_options:
                cmin += self.current - half_options
                cmax += self.current - half_options
            elif self.current >= len(choices) - half_options:
                cmin += len(choices) - MAX_OPTIONS_DISPLAYED_AT_ONCE
                cmax += len(choices)

            cchoices = choices[cmin:cmax]
        else:
            cchoices = choices

        ending_milestone = max(len(choices) - half_options, half_options + 1)
        is_in_beginning = self.current <= half_options
        is_in_middle = half_options < self.current < ending_milestone
        is_in_end = self.current >= ending_milestone

        for index, choice in enumerate(cchoices):
            end_index = ending_milestone + index - half_options - 1
            if (
                (is_in_middle and index == half_options)
                or (is_in_beginning and index == self.current)
                or (is_in_end and end_index == self.current)
            ):

                color = self.theme.List.selection_color
                symbol = "+" if choice == GLOBAL_OTHER_CHOICE else self.theme.List.selection_cursor
            else:
                color = self.theme.List.unselected_color
                symbol = " "
            yield choice, symbol, color

    def process_input(self, pressed):
        question = self.question
        if pressed == key.UP:
            if question.carousel and self.current == 0:
                self.current = len(question.choices) - 1
            else:
                self.current = max(0, self.current - 1)
            return
        if pressed == key.DOWN:
            if question.carousel and self.current == len(question.choices) - 1:
                self.current = 0
            else:
                self.current = min(len(self.question.choices) - 1, self.current + 1)
            return
        if pressed == key.ENTER:
            value = self.question.choices[self.current]

            if value == GLOBAL_OTHER_CHOICE:
                value = self.other_input()
                if not value:
                    # Clear the print inquirer3.text made, since the user didn't enter anything
                    print(self.terminal.move_up + self.terminal.clear_eol, end="")
                    return

            raise errors.EndOfInput(getattr(value, "value", value))

        if pressed == key.CTRL_C:
            raise KeyboardInterrupt()

    def _current_index(self):
        try:
            return self.question.choices.index(self.question.default)
        except ValueError:
            return 0

    def get_current_value(self):
        try:
            return self.question.choices[self.current]
        except IndexError:
            return ""



import math
import sys

from blessed import Terminal

from inquirer3 import errors
from inquirer3 import events
from inquirer3 import themes
from inquirer3.render.console._checkbox import Checkbox
from inquirer3.render.console._confirm import Confirm
from inquirer3.render.console._editor import Editor
from inquirer3.render.console._list import List
from inquirer3.render.console._password import Password
from inquirer3.render.console._path import Path
from inquirer3.render.console._text import Text


class ConsoleRender:
    def __init__(self, event_generator=None, theme=None, raise_keyboard_interrupt: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._event_gen = event_generator or events.KeyEventGenerator()
        self.terminal = Terminal()
        self._previous_error = None
        self._position = 0
        self._theme = theme or themes.Default()
        self._raise_keyboard_interrupt = raise_keyboard_interrupt

    def render(self, question, answers=None):
        question.answers = answers or {}

        if question.ignore:
            return question.default

        clazz = self.render_factory(question.kind)
        render = clazz(question, terminal=self.terminal, theme=self._theme, show_default=question.show_default)

        try:
            return self._event_loop(render)
        finally:
            print("")

    def _event_loop(self, render):
        try:
            while True:
                self._relocate_and_clear()
                self._print_status_bar()

                self._print_header(render)
                self._print_options(render)

                self._process_input(render)

        except KeyboardInterrupt:
            if self._raise_keyboard_interrupt:
                raise
            self.clear_bottombar()
            print("\n", end="\n" if render.title_inline else "")
            print("Cancelled by user")

        except errors.EndOfInput as e:
            self._go_to_end(render)
            return e.selection

    def _print_status_bar(self):
        if self._previous_error is None:
            self.clear_bottombar()
            return

        self.render_error(self._previous_error)
        self._previous_error = None

    def _print_options(self, render):
        for message, symbol, color in render.get_options():
            if hasattr(message, "decode"):  # python 2
                message = message.decode("utf-8")
            choice = f" {color}{symbol} {message}{self.terminal.normal}"
            if render.question.trim_choices:
                choice = self.trim_str(choice)
            self.print_line(choice)

    def _print_header(self, render):
        header = render.get_header()

        header_prefix_template = "{tq.brackets_color}[{tq.mark_color}?{tq.brackets_color}]{t.normal}"
        header_prefix = header_prefix_template.format(tq=self._theme.Question, t=self.terminal)
        header = f"{header_prefix} {header}"

        default_value = render.question.default
        if default_value and render.show_default:
            header += f" ({self._theme.Question.default_color}{default_value}{self.terminal.normal})"

        header_ending = ": "
        if render.question.trim_header:
            header = self.trim_str(header, extra_after_trimming=header_ending)
        else:
            header += header_ending

        full_header = f"{header}{str(render.get_current_value())}"
        self.print_str(full_header, lf=not render.title_inline)

    def _process_input(self, render):
        try:
            ev = self._event_gen.next()
            if isinstance(ev, events.KeyPressed):
                render.process_input(ev.value)
        except errors.ValidationError as e:
            self._previous_error = e.value
        except errors.EndOfInput as e:
            try:
                render.question.validate(e.selection)
                raise
            except errors.ValidationError as e:
                self._previous_error = render.handle_validation_error(e)

    def _relocate_and_clear(self):
        print("\r" + self._position * self.terminal.move_up, end="")
        self.clear_eos()
        leave_space_for_error = self._position + 1
        print("\n" * leave_space_for_error, end="")
        print(self.terminal.move_up * leave_space_for_error, end="")
        self._position = 0

    def _go_to_end(self, render):
        positions = len(list(render.get_options())) - self._position
        if positions > 0:
            print(self._position * self.terminal.move_down, end="")
        self._position = 0

    def render_error(self, message):
        if message:
            symbol = ">> "
            size = len(symbol) + 1
            length = len(message)
            message = message.rstrip()
            message = message if length + size < self.width else message[: self.width - (size + 3)] + "..."

            self.render_in_bottombar(
                "{t.red}{s}{t.normal}{t.bold}{msg}{t.normal} ".format(msg=message, s=symbol, t=self.terminal)
            )

    def render_in_bottombar(self, message):
        with self.terminal.location(0, self.height - 2):
            self.clear_eos()
            self.print_str(message)

    def clear_bottombar(self):
        with self.terminal.location(0, self.height - 2):
            self.clear_eos()

    @staticmethod
    def render_factory(question_type):
        matrix = {
            "text": Text,
            "editor": Editor,
            "password": Password,
            "confirm": Confirm,
            "list": List,
            "checkbox": Checkbox,
            "path": Path,
        }

        if question_type not in matrix:
            raise errors.UnknownQuestionTypeError()
        return matrix.get(question_type)

    def print_line(self, msg, lf=True):
        self.print_str(msg, lf=lf)

    def print_str(self, msg, lf=False):
        print("\r" + msg, end="\n" if lf else "")
        sys.stdout.flush()

        self._position += math.floor((self.terminal.length(msg) - 1) / self.width)
        if lf:
            self._position += 1

    def clear_eos(self):
        print(self.terminal.clear_eos, end="")

    @property
    def width(self):
        return self.terminal.width or 80

    @property
    def height(self):
        return self.terminal.width or 24

    def trim_str(self, msg: str, extra_after_trimming: str = ""):
        extra_if_long = "..."
        maximum_width = self.width - len(extra_if_long + extra_after_trimming)
        if self.terminal.length(msg) > maximum_width:
            return self.terminal.truncate(msg, maximum_width) + extra_if_long + extra_after_trimming
        return msg + extra_after_trimming



import inquirer3.themes as themes
from inquirer3.render.console import ConsoleRender


def prompt(questions, render=None, answers=None, theme=themes.Default(), raise_keyboard_interrupt=False):
    render = render or ConsoleRender(theme=theme, raise_keyboard_interrupt=raise_keyboard_interrupt)
    answers = answers or {}

    for question in questions:
        answers[question.name] = render.render(question, answers)
    return answers



import collections
import json

from blessed import Terminal

import inquirer3.errors as errors


term = Terminal()


def load_theme_from_json(json_theme):
    """Load a theme from a json.

    Expected format:
        >>> {
        ...     "Question": {
        ...         "mark_color": "yellow",
        ...         "brackets_color": "normal",
        ...         ...
        ...     },
        ...     "List": {
        ...         "selection_color": "bold_blue",
        ...         "selection_cursor": "->"
        ...     }
        ... }

    Color values should be string representing valid blessings.Terminal colors.
    """
    return load_theme_from_dict(json.loads(json_theme))


def load_theme_from_dict(dict_theme):
    """Load a theme from a dict.

    Expected format:
        >>> {
        ...     "Question": {
        ...         "mark_color": "yellow",
        ...         "brackets_color": "normal",
        ...         ...
        ...     },
        ...     "List": {
        ...         "selection_color": "bold_blue",
        ...         "selection_cursor": "->"
        ...     }
        ... }

    Color values should be string representing valid blessings.Terminal colors.
    """
    t = Default()
    for question_type, settings in dict_theme.items():
        if question_type not in vars(t):
            raise errors.ThemeError(
                "Error while parsing theme. Question type " "`{}` not found or not customizable.".format(question_type)
            )

        # calculating fields of namedtuple, hence the filtering
        question_fields = list(filter(lambda x: not x.startswith("_"), vars(getattr(t, question_type))))

        for field, value in settings.items():
            if field not in question_fields:
                raise errors.ThemeError(
                    "Error while parsing theme. Field "
                    "`{}` invalid for question type `{}`".format(field, question_type)
                )
            actual_value = getattr(term, value) or value
            setattr(getattr(t, question_type), field, actual_value)
    return t


class Theme:
    def __init__(self):
        self.Question = collections.namedtuple("question", "mark_color brackets_color default_color")
        self.Editor = collections.namedtuple("editor", "opening_prompt")
        self.Checkbox = collections.namedtuple(
            "common",
            "selection_color selection_icon selected_color unselected_color "
            "selected_icon unselected_icon locked_option_color",
        )
        self.List = collections.namedtuple("List", "selection_color selection_cursor unselected_color")


class Default(Theme):
    def __init__(self):
        super().__init__()
        self.Question.mark_color = term.yellow
        self.Question.brackets_color = term.normal
        self.Question.default_color = term.normal
        self.Editor.opening_prompt_color = term.bright_black
        self.Checkbox.selection_color = term.blue
        self.Checkbox.selection_icon = ">"
        self.Checkbox.selected_icon = "[X]"
        self.Checkbox.selected_color = term.yellow + term.bold
        self.Checkbox.unselected_color = term.normal
        self.Checkbox.unselected_icon = "[ ]"
        self.Checkbox.locked_option_color = term.gray50
        self.List.selection_color = term.blue
        self.List.selection_cursor = ">"
        self.List.unselected_color = term.normal


class GreenPassion(Default):
    def __init__(self):
        super().__init__()
        self.Question.brackets_color = term.bright_green
        self.Checkbox.selection_color = term.bold_black_on_bright_green
        self.Checkbox.selection_icon = "❯"
        self.Checkbox.selected_icon = "◉"
        self.Checkbox.selected_color = term.green
        self.Checkbox.unselected_icon = "◯"
        self.List.selection_color = term.bold_black_on_bright_green
        self.List.selection_cursor = "❯"


class BlueComposure(Default):
    def __init__(self):
        super().__init__()
        self.Question.brackets_color = term.dodgerblue
        self.Question.default_color = term.deepskyblue2
        self.Checkbox.selection_icon = "➤"
        self.Checkbox.selection_color = term.bold_black_on_darkslategray3
        self.Checkbox.selected_icon = "☒"
        self.Checkbox.selected_color = term.cyan3
        self.Checkbox.unselected_icon = "☐"
        self.List.selection_color = term.bold_black_on_darkslategray3
        self.List.selection_cursor = "➤"



from typing import Optional, Iterable, Any, Union, Callable, List

import inquirer3.questions as questions
from inquirer3.render.console import ConsoleRender


def text(
    message: Any,
    default: Optional[str] = None,
    ignore: Union[bool, Callable] = False,
    validate: Union[bool, Callable] = True,
    show_default: Union[bool, Callable] = False,
    trim_header: bool = True,
    autocomplete: Optional[Callable] = None,
    render: Optional[ConsoleRender] = None,
    raise_keyboard_interrupt: bool = False,
) -> str:
    render = render or ConsoleRender(raise_keyboard_interrupt=raise_keyboard_interrupt)
    question = questions.Text(
        name="",
        message=message,
        default=default,
        ignore=ignore,
        validate=validate,
        show_default=show_default,
        autocomplete=autocomplete,
        trim_header=trim_header,
    )
    return render.render(question)


def editor(
    message: Any,
    default: Optional[str] = None,
    ignore: Union[bool, Callable] = False,
    validate: Union[bool, Callable] = True,
    show_default: Union[bool, Callable] = False,
    trim_header: bool = True,
    autocomplete: Optional[Callable] = None,
    render: Optional[ConsoleRender] = None,
    raise_keyboard_interrupt: bool = False,
) -> str:
    render = render or ConsoleRender(raise_keyboard_interrupt=raise_keyboard_interrupt)
    question = questions.Editor(
        name="",
        message=message,
        default=default,
        ignore=ignore,
        validate=validate,
        show_default=show_default,
        autocomplete=autocomplete,
        trim_header=trim_header,
    )
    return render.render(question)


def password(
    message: Any,
    echo: str = "*",
    default: Any = None,
    ignore: Union[bool, Callable] = False,
    validate: Union[bool, Callable] = True,
    show_default: Union[bool, Callable] = False,
    trim_header: Union[bool, Callable] = True,
    render: Optional[ConsoleRender] = None,
    raise_keyboard_interrupt: bool = False,
) -> str:
    render = render or ConsoleRender(raise_keyboard_interrupt=raise_keyboard_interrupt)
    question = questions.Password(
        name="",
        message=message,
        echo=echo,
        default=default,
        ignore=ignore,
        validate=validate,
        show_default=show_default,
        trim_header=trim_header,
    )
    return render.render(question)


def confirm(
    message: Any = "",
    default: Any = False,
    ignore: Union[bool, Callable] = False,
    validate: Union[bool, Callable] = True,
    trim_header: Union[bool, Callable] = True,
    render: Optional[ConsoleRender] = None,
    raise_keyboard_interrupt: bool = False,
) -> str:
    render = render or ConsoleRender(raise_keyboard_interrupt=raise_keyboard_interrupt)
    question = questions.Confirm(
        name="",
        message=message,
        default=default,
        ignore=ignore,
        validate=validate,
        trim_header=trim_header,
    )
    return render.render(question)


def list_input(
    choices: Iterable,
    message: Any = "",
    default: Any = None,
    ignore: Union[bool, Callable] = False,
    validate: Union[bool, Callable] = True,
    show_default: Union[bool, Callable] = False,
    other: bool = False,
    trim_header: Union[bool, Callable] = True,
    carousel: bool = False,
    autocomplete: Optional[Callable] = None,
    trim_choices: bool = False,
    render: Optional[ConsoleRender] = None,
    raise_keyboard_interrupt: bool = False,
) -> str:
    render = render or ConsoleRender(raise_keyboard_interrupt=raise_keyboard_interrupt)
    question = questions.List(
        name="",
        choices=choices,
        message=message,
        default=default,
        ignore=ignore,
        validate=validate,
        show_default=show_default,
        other=other,
        trim_header=trim_header,
        carousel=carousel,
        autocomplete=autocomplete,
        trim_choices=trim_choices,
    )
    return render.render(question)


def checkbox(
    choices: Iterable,
    message: Any = "",
    default: Any = None,
    ignore: Union[bool, Callable] = False,
    validate: Union[bool, Callable] = True,
    show_default: Union[bool, Callable] = False,
    other: bool = False,
    trim_header: Union[bool, Callable] = True,
    locked: Optional[List] = None,
    carousel: bool = False,
    autocomplete: Optional[Callable] = None,
    trim_choices: bool = False,
    render: Optional[ConsoleRender] = None,
    raise_keyboard_interrupt: bool = False,
) -> str:
    render = render or ConsoleRender(raise_keyboard_interrupt=raise_keyboard_interrupt)
    question = questions.Checkbox(
        name="",
        choices=choices,
        message=message,
        default=default,
        ignore=ignore,
        validate=validate,
        show_default=show_default,
        other=other,
        trim_header=trim_header,
        locked=locked,
        carousel=carousel,
        autocomplete=autocomplete,
        trim_choices=trim_choices,
    )
    return render.render(question)


def path(
    message: Any = "",
    default: Any = None,
    ignore: Union[bool, Callable] = False,
    validate: Union[bool, Callable] = True,
    show_default: Union[bool, Callable] = False,
    trim_header: Union[bool, Callable] = True,
    autocomplete: Optional[Callable] = None,
    path_type: str = "any",
    exists: Optional[bool] = None,
    normalize_to_absolute_path: bool = False,
    render: Optional[ConsoleRender] = None,
    raise_keyboard_interrupt: bool = False,
) -> str:
    render = render or ConsoleRender(raise_keyboard_interrupt=raise_keyboard_interrupt)
    question = questions.Path(
        name="",
        message=message,
        default=default,
        ignore=ignore,
        validate=validate,
        show_default=show_default,
        trim_header=trim_header,
        autocomplete=autocomplete,
        path_type=path_type,
        exists=exists,
        normalize_to_absolute_path=normalize_to_absolute_path,
    )
    return render.render(question)



from readchar import key

from inquirer3 import errors
from inquirer3.render.console.base import BaseConsoleRender


class Confirm(BaseConsoleRender):
    title_inline = True

    def get_header(self):
        confirm = "(Y/n)" if self.question.default else "(y/N)"
        return f"{self.question.message} {confirm}"

    def process_input(self, pressed):
        if pressed == key.CTRL_C:
            raise KeyboardInterrupt()

        if pressed.lower() == key.ENTER:
            raise errors.EndOfInput(self.question.default)

        if pressed in "yY":
            print(pressed)
            raise errors.EndOfInput(True)
        if pressed in "nN":
            print(pressed)
            raise errors.EndOfInput(False)



from readchar import key

from inquirer3 import errors
from inquirer3.render.console.base import BaseConsoleRender


class Text(BaseConsoleRender):
    title_inline = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current = self.question.default or ""
        self.cursor_offset = 0
        self._autocomplete_state = None

    def get_current_value(self):
        return self.current + (self.terminal.move_left * self.cursor_offset)

    def process_input(self, pressed):
        if pressed == key.CTRL_C:
            raise KeyboardInterrupt()

        if pressed in (key.CR, key.LF, key.ENTER):
            raise errors.EndOfInput(self.current)

        if pressed == key.TAB and self.question.autocomplete:
            if self._autocomplete_state is None:
                self._autocomplete_state = [self.current, 0]

            [text, state] = self._autocomplete_state
            autocomplete = self.question.autocomplete(text, state)
            if isinstance(autocomplete, str):
                self.current = autocomplete
                self._autocomplete_state[1] += 1
            else:
                self._autocomplete_state = None
            return

        self._autocomplete_state = None

        if pressed == key.BACKSPACE:
            if self.current and self.cursor_offset != len(self.current):
                if self.cursor_offset > 0:
                    n = -self.cursor_offset
                    self.current = self.current[: n - 1] + self.current[n:]
                else:
                    self.current = self.current[:-1]
        elif pressed == key.SUPR:
            if self.current and self.cursor_offset:
                n = -self.cursor_offset
                self.cursor_offset -= 1
                if n < -1:
                    self.current = self.current[:n] + self.current[n + 1 :]  # noqa E203
                else:
                    self.current = self.current[:n]
        elif pressed == key.LEFT:
            if self.cursor_offset < len(self.current):
                self.cursor_offset += 1
        elif pressed == key.RIGHT:
            self.cursor_offset = max(self.cursor_offset - 1, 0)
        elif len(pressed) != 1:
            return
        else:
            if self.cursor_offset == 0:
                self.current += pressed
            else:
                n = -self.cursor_offset
                self.current = "".join((self.current[:n], pressed, self.current[n:]))
